#ifdef COPYRIGHT_INFORMATION
#include "gplv3.h"
#endif
/*
 *
 *  Copyright (c) 2004-2011 Edscott Wilson Garcia <edscott@xffm.org>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library;  */

#include <sys/types.h>
#include <sys/wait.h>

#include "rfm.h"

static regex_t supported;
static gboolean regex_compiled = FALSE;

G_MODULE_EXPORT gboolean
svg_supported (void) {
    static int support = 2;
    if(support < 2)
        return (gboolean) support;
    else
        support = 0;
    GSList *l;
    GSList *pix_formats = NULL;
    if(!pix_formats)
        pix_formats = gdk_pixbuf_get_formats ();
    for(l = pix_formats; l; l = l->next) {
        gchar **pix_mimetypes;
        int i;
        GdkPixbufFormat *fmt = l->data;
        pix_mimetypes = gdk_pixbuf_format_get_mime_types (fmt);
        for(i = 0; pix_mimetypes[i]; i++) {
            if(g_ascii_strcasecmp (pix_mimetypes[i], "image/svg") == 0) {
                support = 1;
                break;
            }
        }
        g_strfreev (pix_mimetypes);
    }
    g_slist_free (pix_formats);
    return (gboolean) support;
}

/*cache stuff*/
static int
check_dir (char *path) {
    struct stat st;
    if(stat (path, &st) < 0) {
        if(mkdir (path, 0770) < 0)
            return FALSE;
        return TRUE;
    }
    if(S_ISDIR (st.st_mode)) {
        if(access (path, W_OK) < 0)
            return FALSE;
        return TRUE;
    }
    return FALSE;
}

static gchar *
get_cache_path (void) {
    gchar *cache_path = NULL;
    gchar *cache_dir;
    cache_dir = g_build_filename (USER_DBH_CACHE_DIR, NULL);
    if (!g_file_test(cache_dir, G_FILE_TEST_IS_DIR)){
	if (g_file_test(cache_dir, G_FILE_TEST_EXISTS)){
	    DBG("%s exists and is not a directory!\n", cache_dir);
	    g_free (cache_dir);
	    return NULL;
	}
	if (g_mkdir_with_parents(cache_dir, 0750) < 0){
	    DBG("g_mkdir_with_parents(%s): %s\n", cache_dir, strerror(errno));
	    g_free (cache_dir);
	    return NULL;
	}
    }

    if(!check_dir (cache_dir)) {
	g_error("!check_dir (%s)\n", cache_dir);
        g_free (cache_dir);
        return NULL;
    }

    cache_path = g_strdup_printf ("%s%cGTK.cache.dbh", cache_dir, G_DIR_SEPARATOR);
    g_free (cache_dir);
    NOOP ("ICON: using cache: %s\n", cache_path);
    return cache_path;
}


static gchar *
get_tmp_cache_path (void) {

    gchar *cache_path = get_cache_path();
    if (!cache_path) return NULL;
    gchar *tmp_cache_path = g_strdup_printf("%s-0x%x-%d", cache_path, GPOINTER_TO_INT(g_thread_self()), getpid());
    g_free(cache_path);
    return tmp_cache_path;
}


static long long
recurse_basedir_sum (const gchar * dir) {
    long long sum = 0;
    const char *file;
    GDir *gdir;
    NOOP ("ICON: recurse_basedir_sum(%s)\n", dir);
    if((gdir = g_dir_open (dir, 0, NULL)) != NULL) {
        while((file = g_dir_read_name (gdir))) {
            gchar *path = g_build_filename (dir, file, NULL);
            if(g_file_test (path, G_FILE_TEST_IS_SYMLINK)) {
                g_free (path);
                continue;
            }
            if(g_file_test (path, G_FILE_TEST_IS_DIR)) {
                struct stat st;
                /* recurse */
                sum += recurse_basedir_sum (path);
                if(stat (path, &st) == 0) {
                    sum += st.st_mtime;
                    sum += st.st_dev;
                }
            }
            g_free (path);
        }
        g_dir_close (gdir);
    }
    return sum;
}


/* must get info for all directories within */
static off_t
get_basedir_sum (void) {
    gint build=0;
    off_t basedir_sum = 0;

    GtkIconTheme *icon_theme = gtk_icon_theme_get_default ();
    gchar **pathv=NULL;
    gtk_icon_theme_get_search_path (icon_theme, &pathv, NULL);
    gchar **p=pathv;
    for (;p && *p; p++) {
	const gchar *basedir = *p;
	struct stat st;
	if(stat (basedir, &st) == 0) {
	    NOOP ("ICON: stat ok\n");
	    gint sum = st.st_mtime + st.st_size + st.st_mode + st.st_nlink + st.st_uid + st.st_gid;
	    basedir_sum += sum;
	}
	basedir_sum += recurse_basedir_sum (basedir);
	NOOP ("ICON-DBH: basedir summing %s (%ld %ld)\n", basedir, (long)st.st_mtime, (long)st.st_dev);
	basedir_sum += build;
    }
    g_strfreev(pathv);
    return basedir_sum;
}

static const gchar *
get_supported_regex (void
    ) {
    GSList *pix_formats,
     *l;
    gchar **pix_extensions,
    **p;
    static gchar *reg = NULL,
        *r = NULL;
    /* check SVG format support */
    if((pix_formats = gdk_pixbuf_get_formats ()) != NULL) {
        for(l = pix_formats; l; l = l->next) {
            GdkPixbufFormat *fmt = l->data;
            pix_extensions = gdk_pixbuf_format_get_extensions (fmt);
            for(p = pix_extensions; *p; p++) {
                NOOP ("ICON: supported=%s\n", *p);
                if(reg) {
                    g_free (r);
                    r = reg;
                    reg = g_strconcat (r, "|", *p, NULL);
                } else
                    reg = g_strdup (*p);
            }
            g_strfreev (pix_extensions);
        }
        g_slist_free (pix_formats);
    }
    if(!reg)
        return "\\.(png|xpm)$)";
    g_free (r);
    r = g_strconcat ("\\.(", reg, ")$", NULL);
    g_free (reg);
    reg = NULL;

    NOOP ("ICON: regex=%s\n", r);
    return (const gchar *)r;
}

static void
save_cache_info (const gchar * version) {
    FILE *info;

    gchar *cache_file = get_cache_path ();
    if (!cache_file) return;
    gchar *cache_info_path = g_strconcat (cache_file, ".info", NULL);
    cache_info_t cache_info;
    NOOP ("0x%x: saving %s\n", GPOINTER_TO_INT(g_thread_self()), cache_info_path);

    strncpy (cache_info.supported_regex, get_supported_regex (), 255);
    cache_info.supported_regex[255] = 0;
    cache_info.basedir_sum = get_basedir_sum ();


    strncpy (cache_info.version, version, 63);
    cache_info.version[63]=0;

    NOOP ("ICON: saving cache_info.version=%s\n", version);
    info = fopen (cache_info_path, "wb");
    NOOP(stderr, "..... writing %s: %s %s \n", cache_info_path, cache_info.version,version);
    if(info) {
        if(fwrite (&cache_info, sizeof (cache_info_t), 1, info) < 1) {
            DBG ("cannot write to %s\n", cache_info_path);
        }
        fclose (info);
    } else
        DBG ("cannot write to %s\n", cache_info_path);
    NOOP ("ICON: wrote basedir sum is %lld\n", cache_info.basedir_sum);
    g_free (cache_info_path);
    g_free (cache_file);
}

/* criteria for cache regeneration:
 * 1- does not exist
 * 2- one of the base_dirs has been modified
 * 3- supported regexp has changed.
 * 4- prefix has changed
 *
 * */
static gboolean
compare_cache_info (void) {
    FILE *info;
    gchar *cache_info_path;
    cache_info_t cache_info;
    cache_info_t  disk_cache_info;
    gint records;

    gchar *cache_file = get_cache_path ();
    if (!cache_file) {
	DBG("no cache file string\n");
        return TRUE;
    }
    NOOP (stderr, "0x%x: comparing cache info \n", GPOINTER_TO_INT(g_thread_self()) );

    if(!g_file_test (cache_file, G_FILE_TEST_EXISTS)) {
        NOOP (stderr, "ICON: %s not exists\n", cache_file);
        g_free (cache_file);
        return FALSE;
    }

    cache_info_path = g_strconcat (cache_file, ".info", NULL);
    g_free (cache_file);
    if((info = fopen (cache_info_path, "rb")) == NULL) {
        NOOP (stderr, "ICON: %s not exists\n", cache_info_path);
        g_free (cache_info_path);
        return FALSE;
    }

    records = fread (&disk_cache_info, sizeof (cache_info_t), 1, info);
    NOOP(stderr, ".... compare cache info: reading %s .....%s\n", cache_info_path, disk_cache_info.version);
    g_free (cache_info_path);
    fclose (info);
    if(records < 1) {
        NOOP (stderr, "records < 1\n");
	return FALSE;
    }

    gchar *version=NULL;
    GtkSettings *settings = gtk_settings_get_default();
    g_object_get( G_OBJECT(settings), 
	    "gtk-icon-theme-name", &version,
	    NULL);    
    if (strlen(version)>64) version[63]=0;

    NOOP ("ICON: read basedir cache sum is %lld\n", disk_cache_info.basedir_sum);

    if(version && strncmp (disk_cache_info.version, version, 64) != 0) {
        NOOP (stderr, "Icontheme version cahnge: %s --> %s\n", disk_cache_info.version, version);
	g_free(version);
        return FALSE;
    }
    g_free(version);


    cache_info.basedir_sum = get_basedir_sum ();
    strncpy (cache_info.supported_regex, get_supported_regex (), 255);
    cache_info.supported_regex[255] = 0;
    NOOP ("ICON: cache sum is %lld\n", cache_info.basedir_sum);
    if(cache_info.basedir_sum != disk_cache_info.basedir_sum) {
        NOOP
            (stderr, "ICON: cache_info.basedir_sum(%lld) != disk_cache_info.basedir_sum(%lld)\n",
             cache_info.basedir_sum, disk_cache_info.basedir_sum);
        return FALSE;
    }

    if(disk_cache_info.supported_regex && 
	    strlen(disk_cache_info.supported_regex) &&
	    strcmp (cache_info.supported_regex, disk_cache_info.supported_regex)) {
        NOOP (stderr, "ICON: cache_info.supported_regex != disk_cache_info.supported_regex\n");
        return FALSE;
    }
    return TRUE;
}

#if 10
static 
gchar *
add_gtk_icons(void){
    rfm_global_t *rfm_global_p = rfm_global();
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	g_error("add_gtk_icons should be main thread run...\n");
    }
    NOOP(stderr, "------>  add_gtk_icons\n");
    static gboolean adding=FALSE;
    if (adding) return NULL;
    adding=TRUE;
    gchar *tmp_cache_file = get_tmp_cache_path ();
    NOOP (stderr, "ICON-DBH: gtk cache--> %s\n", tmp_cache_file);

    unsigned char keylength=11;
    gchar *directory = g_path_get_dirname(tmp_cache_file);
    if (!g_file_test(directory, G_FILE_TEST_IS_DIR)){
        g_mkdir_with_parents(directory, 0700);
    }
    g_free(directory);
    DBHashTable *cache = dbh_new(tmp_cache_file, &keylength, DBH_CREATE);
    if(cache == NULL) {
	DBG("* Could not create GTK icon module cache: %s\n", tmp_cache_file);
	g_free (tmp_cache_file);
	return NULL;;
    }
    NOOP (stderr, "ICON-DBH: empty cache created...\n");


    
    GtkIconInfo *icon_info;
    const gchar *path;
    GtkIconTheme *icon_theme = gtk_icon_theme_get_default ();
    GList *list=NULL;
    if (icon_theme) list = gtk_icon_theme_list_icons (icon_theme, NULL);
    GList *tmp=list;
    TRACE( "Creating cache for GTK icontheme...\n");
    for (;tmp && tmp->data; tmp=tmp->next){
	gchar *icon_name = tmp->data;
	// Get path for size 128 or 48.

	icon_info =
	    gtk_icon_theme_lookup_icon (icon_theme,
		    icon_name, 128, GTK_ICON_LOOKUP_GENERIC_FALLBACK);
	if (!icon_info){
	    icon_info =
		gtk_icon_theme_lookup_icon (icon_theme,
		    icon_name, 48, GTK_ICON_LOOKUP_GENERIC_FALLBACK);
	}
	if (!icon_info){
	    NOOP(stderr, "no theme icon for %s\n", icon_name);
	    continue;
	}

	path = gtk_icon_info_get_filename (icon_info);
	if (!path){
	    NOOP(stderr, "Path for %s is NULL\n", icon_name);
	    continue;
	}

	// Put absolute paths into icon hash...
	gchar *hash_key = rfm_get_hash_key(icon_name, 0);
        memset (DBH_KEY (cache), 0, DBH_KEYLENGTH (cache));
        sprintf ((gchar *)DBH_KEY (cache), "%10s", hash_key);
	NOOP(stderr, "inserting into dbh cache: %s (%s)-> %s\n", icon_name, hash_key, path);
	g_free(hash_key);
        if(dbh_load (cache)) {
            NOOP (stderr, "gtk icontheme: add2cache, %s --> %s already in cache! \ncache value:%s\nignoring %s\n", (char *)DBH_KEY (cache), hash_key, (gchar *) cache->data, path);
        } else {
            strcpy ((gchar *) cache->data, path);
            dbh_set_recordsize (cache, strlen (path) + 1);
            if(!dbh_update (cache)) {
                NOOP (stderr, "ICON-DBH: !dbh_update failed!\n");
            }

        }
	// Free no longer useful gtk_icon_info.
	// Beware that const gchar *path belongs to gtk_icon_info
#if GTK_MAJOR_VERSION==3 && GTK_MINOR_VERSION>=8
	if (icon_info)g_object_unref(G_OBJECT(icon_info));
#else
	if (icon_info) gtk_icon_info_free (icon_info); 
#endif
	NOOP("icon: %s\n", (gchar *)tmp->data);
	g_free(tmp->data);
    }
    g_list_free(list);   

    // Optimize speed by making logical structure match physical structure:
    dbh_regen_sweep(&cache);
    dbh_close (cache);

    gchar *current_version=NULL;
    GtkSettings *settings = gtk_settings_get_default();
    g_object_get( G_OBJECT(settings), 
	    "gtk-icon-theme-name", &current_version,
	    NULL);    
    save_cache_info (current_version);
    g_free(current_version);
     
    TRACE( "GTK icontheme cache is now written.\n");
    adding=FALSE;
    return tmp_cache_file;
}

static void *
create_cache_f(void *data){
    rfm_global_t *rfm_global_p = rfm_global();
    /* create empty hash file */
    if (rfm_global_p){
        g_mutex_lock(rfm_global_p->status_mutex);
        gint status = rfm_global_p->status;
        g_mutex_unlock(rfm_global_p->status_mutex);
        if (status == STATUS_EXIT){
            return NULL;
        }
    }
    NOOP (stderr, "----> Generating GTK icon-module cache. This may take a bit...\n");
    gchar *tmp_cache_file = add_gtk_icons ();
    if (!tmp_cache_file){
	DBG("** Could not create GTK icon module cache\n" );
	return NULL;
    }
    if (rfm_global_p){
        g_mutex_lock(rfm_global_p->status_mutex);
        gint status = rfm_global_p->status;
        g_mutex_unlock(rfm_global_p->status_mutex);
        if (status == STATUS_EXIT) {
            unlink(tmp_cache_file);
            g_free(tmp_cache_file);
            return NULL;
        }
    }
    gchar *cache_file = get_cache_path ();
    if (!cache_file) {
	unlink(tmp_cache_file);
	g_free (tmp_cache_file);
	return NULL;
    }
    // This add non themed icons.
    //add_theme_directories();
    NOOP (stderr, "-----> GTK cache complete:\n%s --> %s\n", tmp_cache_file, cache_file);
    if (rename(tmp_cache_file, cache_file) < 0) {
	DBG("rename(%s, %s): %s\n", 
		tmp_cache_file, cache_file, strerror(errno));
    }
    g_free (tmp_cache_file);
    g_free (cache_file);
    return GINT_TO_POINTER(1);
}

G_MODULE_EXPORT
gboolean
create_cache(){
    void *retval = rfm_context_function(create_cache_f, NULL);
    return GPOINTER_TO_INT(retval);
}


static void *
create_new_gtk_cache(void *data) {
    // data != NULL means we are in background thread.
    // data == NULL means the main thread is here, updating
    //         the cache before any icon is displayed.

    if (!data) {
	create_cache();
	return NULL;
    }

    static gchar *icon_theme_name;
    if (getenv("RFM_USE_GTK_ICON_THEME"))
	icon_theme_name = g_strdup(getenv("RFM_USE_GTK_ICON_THEME"));
    else icon_theme_name = g_strdup("");
    
    // If running by thread, give main thread a head start.
    gint wait_period=GPOINTER_TO_INT(data);
    if (wait_period > 5){
	DBG("wait_period > 5 is dumb.\n");
	wait_period=5;
    }
    NOOP("0x%x waiting %d seconds before creating GTK icon cache\n", 
	    GPOINTER_TO_INT(g_thread_self()), wait_period);

    sleep(wait_period);

    // This will monitor changes in icontheme name
    rfm_global_t *rfm_global_p = rfm_global();
    do {
	if (rfm_global_p){
            g_mutex_lock(rfm_global_p->status_mutex);
            gint status = rfm_global_p->status;
            g_mutex_unlock(rfm_global_p->status_mutex);
            if(status == STATUS_EXIT) return FALSE;
	}
	// Loop wait. Use rfm_threadwait() for stress tests...
	//rfm_threadwait();
	sleep(2);
	if (rfm_global_p){
            g_mutex_lock(rfm_global_p->status_mutex);
            gint status = rfm_global_p->status;
            g_mutex_unlock(rfm_global_p->status_mutex);
            if(status == STATUS_EXIT) return FALSE;
	}
	const gchar *current_icon_theme = getenv("RFM_USE_GTK_ICON_THEME");
	if (!current_icon_theme || strlen(current_icon_theme)==0){
	    continue;
	}
	if (strcmp(current_icon_theme, icon_theme_name)==0) continue;
	g_free(icon_theme_name);

	icon_theme_name = g_strdup(current_icon_theme);
    } while (1);
    NOOP(stderr, "--Ending gtk icon theme thread (will not monitor changes in theme).\n");
	    
    return NULL;
}
#endif

static int
open_theme (void) {
    // If there is a version mismatch, then the cache will not
    // work with this version of Rodent and must be regenerated
    // before the main thread can continue.
    //
    if (gtk_icon_theme_get_default ()) {
	// Main thread should wait for cache to be generated.
	// If cache is up to date, then this function call should
	// return ipso facto.
	//
	// We pass NULL here so that function will not
	// enter endless loop to monitor gtk icon theme. 
	// Data parameter is also the nice wait period, in
	// seconds.
	if (!compare_cache_info ()){
	    create_new_gtk_cache(NULL);
	} 
	// fire up gtk theme monitor.
	NOOP ("** background Starting GTK theme monitor\n");
	THREAD_CREATE( create_new_gtk_cache, GINT_TO_POINTER(5), "create_new_gtk_cache");
    }
    return 1;
}

static gchar *
get_dbh_pixmap_path (const gchar * key) {
    if (!key) return NULL;
    NOOP ("get_dbh_pixmap_path(GTK): icon lookup: %s\n", key);
    gchar *cache_file = get_cache_path ();
    if(!cache_file) {
        DBG ("ICON: unable to get cache filename for GTK\n");
        return NULL;
    }
    gchar *cache_hit = NULL;
    DBHashTable *cache = NULL;

    // 
    // We could keep the DBH table open and use thread
    // control, or open and close DBH table in read only
    // mode on each access. The latter is faster, since
    // the table is in the OS filesystem cache.
    //
    if((cache = dbh_new (cache_file, NULL, DBH_READ_ONLY)) == NULL) {
	// This warning will popup once for each icon on screen if the
	// GTK icon cache has not been generated and we do an icontheme
	// switch. Otherwise, it would be terribly wrong... 
        DBG ("This is terribly wrong: cannot open icon cache file %s\n", cache_file);
	g_free(cache_file);
        return NULL;
    } 
    g_free(cache_file);
    gchar *hash_key=rfm_get_hash_key (key, 0);
    memset (DBH_KEY (cache), 0, DBH_KEYLENGTH (cache));
    sprintf ((char *)DBH_KEY (cache), "%10s", hash_key);
    g_free(hash_key);
    if(!dbh_load (cache)) {
	NOOP (" icon not found: %s (%s)\n", 
		 key, (char *)DBH_KEY (cache));
	dbh_close (cache);
	return NULL;
    }
    cache_hit = g_strdup ((gchar *) cache->data);
    if (g_file_test (cache_hit, G_FILE_TEST_EXISTS)){
	NOOP ("GTK: cache hit: %s -> %s\n", key, cache_hit);
    } else {
	NOOP ("GTK: cache value invalid: %s -> %s\n", key, cache_hit);
	g_free(cache_hit);
	cache_hit=NULL;
    }
    dbh_close (cache);
    return cache_hit;
}


static 
gint
load_rodent_theme (void) {
    // This is to create a cache for full paths for icons.
    if(open_theme () < 0) {
        DBG ("load_rodent_theme(): cannot load theme\n");
        return -1;
    }

    // This is to create a g_hash which will associate mimetype with an icon
    // the mimetype/icon association hash is created by parsing the xml file.
    gchar *mimefile = mime_icon_get_local_xml_file ();
    NOOP ("load_rodent_theme(): looking for: %s\n", (mimefile ? mimefile : "null"));

    /* test for default mimefile  */
    if(!mimefile || !g_file_test (mimefile, G_FILE_TEST_EXISTS)) {
       NOOP ("%s: not found. \n", (mimefile ? mimefile : "null"));
       g_free (mimefile);
       mimefile = mime_icon_get_global_xml_file ();
       NOOP ("Now looking for: %s\n", (mimefile ? mimefile : "null"));
    }

    if(!mimefile || !g_file_test (mimefile, G_FILE_TEST_EXISTS)) {
        DBG ("No system wide mime file found: %s\n", (mimefile ? mimefile : "null"));
        g_free (mimefile);
        mimefile = NULL;
	return FALSE;
    }

    gboolean result = create_icon_hash (mimefile);
    g_free (mimefile);
    if(!result) {
	DBG ("cannot create basename_hash from mimefile:%s!\n", (mimefile ? mimefile : "null"));
    }
    return result;
}


